import java.io.*;
import java.net.*;

public abstract class FileManager extends Thread implements Messenger
{
	private static final int BUFFER_INCREMENT=8*1024;
	private static final int BUFFER_MAX=2*1024*1024;

	protected static Class [] ARGV={(new String[1]).getClass(),(new String()).getClass()};

	public static final int FM_DOWNLOAD_OK=1;
	public static final int FM_DOWNLOAD_CANCEL=2;
	public static final int FM_UPLOAD_OK=1;
	public static final int FM_UPLOAD_CANCEL=2;
	public static final int FM_UPLOAD_INVALID=3;

	protected boolean not_exit;
	protected byte [] buffer;
	protected DataInputStream peer_in;
	protected DataOutputStream peer_out;
	protected File local_pwd;
	protected Socket socket;
	protected int r_type,cont;
	protected String reply;
	protected boolean no_to_all,not_yes_all;
	protected String [] srcTrg;

	public FileManager(Socket socket,File local_pwd)
	{
		int tmp;
		try {
			peer_out=new DataOutputStream(socket.getOutputStream());
			peer_in=new DataInputStream(socket.getInputStream());
			this.local_pwd=local_pwd;
			this.socket=socket;
			not_exit=true;
			tmp=BUFFER_MAX/socket.getSendBufferSize();
			tmp= tmp > 0 ? tmp*socket.getSendBufferSize() : BUFFER_MAX;
			buffer=new byte[tmp];
			no_to_all=false;
			not_yes_all=true;
			srcTrg=new String[2];
		} catch (Exception e) {
			not_exit=false;
		}
	}

	protected void readCommand()
	{
		try {
			r_type=readByte();
			if (r_type != 100) {
				reply=readLine();
			}
			cont=readByte();
		} catch (SocketException se) {
			r_type=-1;
		}
	}

	public void userMessage(byte command,String message,byte cont)
	{
		if (not_exit) {
			try {
				writeByte(command);
				if (command != 100) {
					writeBytes(message.getBytes());
					writeByte('\n');
				}
				writeByte(cont);
			} catch (SocketException se) { }
		}
	}

	private void innerMv(String src,String dst)
	{
		try {
			File srcF=getAbsoluteFile(src);
			File dstF=getAbsoluteFile(dst);
			if (dstF.isDirectory()) {
				dstF=getAbsoluteFile(dstF.getAbsolutePath()+File.separatorChar+srcF.getName());
			}
					
			if (!srcF.renameTo(dstF)) {
				userMessage((byte)0,"Hata: "+src+"'dan "+src+"'ya tasima islemi basarisiz oldu",(byte)1);
			}
		} catch (Exception e) {
			userMessage((byte)0,"Beklenmedik Durum: "+e.getMessage(),(byte)1);
		}
	}

	public void mv(String [] cmd,String cmdStr)
	{
		try {
			if (cmd.length >= 3) {
				if (cmd.length > 3) {
					File tmp=getAbsoluteFile(cmd[cmd.length-1]);
					if (!tmp.isDirectory()) {
						userMessage((byte)0,"Hata: Dizin olmayan bir hedefe tasima islemi"+tmp.getName(),(byte)0);
						return;
					}
				}
				for (int i=1, j=cmd.length-1; i < j; i++) {
					innerMv(cmd[i],cmd[j]);
				}
				userMessage((byte)100,null,(byte)0);
			} else {
				userMessage((byte)0,"Hata: Yetersiz sayida degistirge",(byte)0);
			}
		} catch (Exception e) {
			userMessage((byte)0,"Beklenmedik Durum: "+e.getMessage(),(byte)0);
		}
	}

	public void cd(String [] cmd,String cmdStr)
	{
		if (cmd.length == 2) {
			try {
				File f=getAbsoluteFile(cmd[1]);
				if (f.isDirectory()) {
					local_pwd=f;
					userMessage((byte)5,f.getAbsolutePath(),(byte)0);
				} else {
					userMessage((byte)0,"Hata : '"+cmd[1]+"' bir dizin degil",(byte)0);
				}
			} catch(Exception e) {
				userMessage((byte)0,"Beklenmedik Durum: "+e.getMessage(),(byte)0);
			}
		} else if (cmd.length == 1) {
			userMessage((byte)0,"cd : Dizin ismi belirtilmedi",(byte)0);
		} else {
			userMessage((byte)0,"cd: Birden fazla dizin ismi",(byte)0);
		}
	}

	private void innerLs(String dirS)
	{
		try {
			File dir=getAbsoluteFile(dirS);
			if (dir.exists()) {
				if (dir.isDirectory()) {
					File [] files=dir.listFiles();
					if (files != null) {
						userMessage((byte)1,"Dizin "+dir.getAbsolutePath()+" Icerigi:",(byte)1);
						for (int i=0; not_exit && i < files.length; i++) {
							try {
								userMessage((byte)1,files[i].getName()+'\t'+(files[i].isDirectory() ? "<DIZIN>" : files[i].isFile() ? "<DOSYA>" : "<Bilinmiyor>")+'\t'+files[i].length(),(byte)1);
							} catch(Exception e) {
								userMessage((byte)0,"Beklenmedik Durum: "+e.getMessage(),(byte)1);
							}
						}
					}
				} else {
					userMessage((byte)1,dir.getName()+'\t'+(dir.isFile() ? "<DOSYA>" : "<Bilinmiyor>")+'\t'+dir.length(),(byte)1);
				}
			} else {
				userMessage((byte)0,"Hata: "+dir.getName()+" mevcut degil",(byte)1);
			}
		} catch(Exception e2) {
			userMessage((byte)0,"Beklenmedik Durum: "+e2.getMessage(),(byte)1);
		}
	}

	public void ls(String [] cmd,String cmdStr)
	{
		if (cmd.length == 1) {
			innerLs(local_pwd.getAbsolutePath());
		} else {
			for (int i=1; not_exit && i < cmd.length; i++) {
				innerLs(cmd[i]);
			}
		}
		userMessage((byte)100,null,(byte)0);
	}

	public void dir(String [] cmd,String cmdStr)
	{
		ls(cmd,cmdStr);
	}

	public void pwd(String [] cmd,String cmdStr)
	{
		try {
			userMessage((byte)2,local_pwd.getAbsolutePath(),(byte)0);
		} catch(Exception e) {
			userMessage((byte)0,"Beklenmedik Durum: "+e.getMessage(),(byte)0);
		}
	}

	public void mkdir(String [] cmd,String cmdStr)
	{
		if (cmd.length >= 2) {
			for (int i=1; not_exit && i < cmd.length; i++) {
				File f=null;
				try {
					f=getAbsoluteFile(cmd[i]);
					if (!f.mkdirs()) {
						userMessage((byte)0,f.getAbsolutePath()+" dizini yaratilmadi",(byte)1);
					}
				} catch(Exception e) {
					userMessage((byte)0,"Beklenmedik Durum:"+f.getAbsolutePath()+":"+e.getMessage(),(byte)1);
				}
			}
			userMessage((byte)100,null,(byte)0);
		} else {
			userMessage((byte)0,"Hata: Yaratilmasi istenen dizin ismi belirtilmedi",(byte)0);
		}
	}

	public void mk(String cmd[],String cmdStr)
	{
		mkdir(cmd,cmdStr);
	}

	public void rm(String cmd[],String cmdStr)
	{
		if (cmd.length >= 2) {
			for (int i=1; not_exit && i < cmd.length; i++) {
				try {
					File f=getAbsoluteFile(cmd[i]);
					if (f.isFile()) {
						if (!f.delete()) {
							userMessage((byte)0,cmd[i]+" silinemedi",(byte)1);
						}
					} else {
						userMessage((byte)0,cmd[i]+" bir dosya degil",(byte)1);
					}
				} catch(Exception e) {
					userMessage((byte)0,"Beklenmedik Durum:"+cmd[i]+":"+e.getMessage(),(byte)1);
				}
			}
			userMessage((byte)100,null,(byte)0);
		} else {
			userMessage((byte)0,"Hata: Silinmesi istenen dosya ismi belirtilmedi",(byte)0);
		}
	}

	public void rmdir(String cmd[],String cmdStr)
	{
		if (cmd.length >= 2) {
			for (int i=1; not_exit && i < cmd.length; i++) {
				try {
					File f=getAbsoluteFile(cmd[i]);
					if (f.isDirectory()) {
						if (!f.delete()) {
							userMessage((byte)0,cmd[i]+" silinemedi",(byte)1);
						}
					} else {
						userMessage((byte)0,cmd[i]+" bir dizin degil",(byte)1);
					}
				} catch(Exception e) {
					userMessage((byte)0,"Beklenmedik Durum:"+cmd[i]+":"+e.getMessage(),(byte)1);
				}
			}
			userMessage((byte)100,null,(byte)0);
		} else {
			userMessage((byte)0,"Hata: Silinmesi istenenen dizin ismi belirtilmedi",(byte)0);
		}
	}

	protected boolean changeLocalPWD(String pwd)
	{
		try {
			local_pwd=getAbsoluteFile(pwd);
			return true;
		} catch (Exception e) {
			userMessage((byte)0,"Beklenmedik Durum: "+e.getMessage(),(byte)1);
		}
		return false;
	}

	protected void splitSourceTarget(String s,String []r)
	{
		int index=s.indexOf('>');
		if (index >= 0) {
			if (s.length() > index+1) {
				r[0]=s.substring(0,index);
				r[1]=s.substring(index+1);
			} else {
				r[0]=null;
			}
		} else {
			r[0]=s;
			r[1]=s;
		}
	}
	
	private void innerGet(String fname,String dname)
 	{
		try {
			File f=getAbsoluteFile(fname);
			if (f.isFile()) {
				peerMessage((byte)3,dname,(byte)1);
				try {
					switch (uploadFile(f)) {
						case FM_UPLOAD_OK :
							return;
						case FM_UPLOAD_CANCEL :
							peerMessage((byte)0,"Uyari: get "+f.getName()+" iptal edildi",(byte)1);
							return;
						default:
							not_exit=false;
							return;
					}
				} catch (SocketException se) { }
				catch (Exception e) { /* Either FileNotFound or SecurityException or NullPointerException */
					if (readByte() == (byte)1) {
						writeLong((long)-1); /* Terminate upload */
					}
					peerMessage((byte)0,"Beklenmedik durum: "+e.getMessage(),(byte)1);
				}
			} else {
				userMessage((byte)0,"Hata: "+f.getName()+" bir dosya degil.",(byte)1);
			}
		} catch (SocketException se) { }			
		catch (Exception e) {
			peerMessage((byte)0,"Beklenmedik Durum: "+e.getMessage(),(byte)1);
		}
	}

	public void get(String cmd[],String cmdStr)
	{
		if (cmd.length == 2) {
			splitSourceTarget(cmd[1],srcTrg);
			if (srcTrg[0] != null) {
				innerGet(srcTrg[0],srcTrg[1]);
				peerMessage((byte)100,null,(byte)0);
			} else {
				peerMessage((byte)0,"Hata: Bos hedef isim: Durum "+cmd[1]+". Iptal ediliyor",(byte)0);
			}
		} else {
			peerMessage((byte)0,"Hata: Dosya ismi belirtilmedi",(byte)0);
		}
	}

	private boolean mgetInner(String [] files,int i)
	{
		File local_backup=null;
		--i;
		while (not_exit && ++i < files.length) {
			try {
				splitSourceTarget(files[i],srcTrg);
				if (srcTrg[0] == null) {
					peerMessage((byte)0,"Hata: Bos hedef isim: Durum "+files[i]+". Atlaniyor ...",(byte)1);
					continue;
				}
				File f=getAbsoluteFile(srcTrg[0]);
				if (f.isFile()) {
					innerGet(srcTrg[0],srcTrg[1]);
				} else if (f.isDirectory()) {
					peerMessage((byte)7,"",(byte)1);
					String l_pwd=readLine();
					peerMessage((byte)4,srcTrg[1],(byte)1);
					if (readByte() == 1) {
						local_backup=local_pwd;
						local_pwd=f;
						String [] dircont=f.list();
						if (mgetInner(dircont,0)) {
							peerMessage((byte)6,l_pwd,(byte)1);
							if (readByte() != 1) {
								peerMessage((byte)0,"Hata: Istemci kendi yerel dizinini degistermeyi basaramadi. mget oturumu kapaniyor",(byte)1);
								peerMessage((byte)0,"Hata: Yerel dizini !pwd ile guncelleyin...",(byte)1);
								return false;
							}
						} else {
							return false;
						}
					}
				} else {
					peerMessage((byte)0,"Hata: Bilinmeyen dosya tipi '"+f.getName()+"'",(byte)1);
				}
			} catch (SocketException se) {
				return false;
			} catch (Exception e) {
				peerMessage((byte)0,"Beklenmedik Durum: "+e.getMessage(),(byte)1);
			} finally {
				if (local_backup != null) {
					local_pwd=local_backup;
				}
			}
		}
		return true;
	}

	public void mget(String [] cmd,String cmdStr)
	{
		if (cmd.length >= 2) {
			try {
				peerMessage((byte)7,"",(byte)1);
				String l_pwd=readLine();
				mgetInner(cmd,1);
				peerMessage((byte)6,l_pwd,(byte)0);
				if (readByte() != 1) {
					userMessage((byte)0,"Hata: Sunucu ustundeki dizin degistirimi basarisiz oldu.",(byte)0);
				}
			} catch (SocketException se) { }
		} else {
			peerMessage((byte)0,"Hata: Dosya/Dizin belirtilmedi",(byte)0);
		}
	}

	protected File getAbsoluteFile(String name) throws Exception
	{
		File f=new File(name);
		if (!f.isAbsolute()) {
			f=(new File(local_pwd.getAbsolutePath()+File.separatorChar+name)).getCanonicalFile();
		}
		return f;
	}

	protected void writeByte(int response) throws SocketException
	{
		try {
			peer_out.writeByte(response);
			peer_out.flush();
		} catch (IOException io) {
			not_exit=false;
			throw new SocketException();
		}
	}

	protected void writeBytes(byte [] buffer) throws SocketException
	{
		try {
			peer_out.write(buffer,0,buffer.length);
			peer_out.flush();
		} catch (IOException io) {
			not_exit=false;
			throw new SocketException();
		}
	}

	protected int readByte() throws SocketException
	{
		try {
			return peer_in.readByte();
		} catch (Exception e) {
			not_exit=false;
		}
		throw new SocketException();
	}

	private void extendBuffer()
	{
		byte [] new_buffer=new byte[buffer.length+BUFFER_INCREMENT];
		for (int i=0; i <= buffer.length; i++) {
			new_buffer[i]=buffer[i];
		}
		buffer=new_buffer;
	}

	protected String readLine() throws SocketException
	{
		try {
			int ind,ch;
			ind=0;
			while ((ch=peer_in.read()) != '\n') {
				buffer[ind++]=(byte)ch;
				if (ind == buffer.length) {
					extendBuffer();
				}
			}
			return new String(buffer,0,ind);
		} catch (Exception e) {
			not_exit=false;
		}
		throw new SocketException();
	}

	protected String writeLine(byte [] line) throws SocketException
	{
		try {
			peer_out.write(line);
			peer_out.writeByte('\n');
		} catch (Exception e) {
			not_exit=false;
		}
		throw new SocketException();
	}
	protected void writeLong(long v) throws SocketException
	{
		try {
			peer_out.writeLong(v);
			peer_out.flush();
		} catch (IOException io) {
			not_exit=false;
			throw new SocketException();
		}
	}

	protected void writeInt(int v) throws SocketException
	{
		try {
			peer_out.writeInt(v);
			peer_out.flush();
		} catch (IOException io) {
			not_exit=false;
			throw new SocketException();
		}
	}

	protected int readInt() throws SocketException
	{
		try {
			return peer_in.readInt();
		} catch (Exception e) {
			not_exit=false;
			throw new SocketException();
		}
	}

	protected long readLong() throws SocketException
	{
		try {
			return peer_in.readLong();
		} catch (Exception e) {
			not_exit=false;
			throw new SocketException();
		}
	}

	protected int read(int size) throws SocketException
	{
		try {
			return peer_in.read(buffer,0,size);
		} catch (IOException io) {
			not_exit=false;
			throw new SocketException();
		}
	}

	protected int read(int off,int size) throws SocketException
	{
		try {
			return peer_in.read(buffer,off,size);
		} catch (IOException io) {
			not_exit=false;
			throw new SocketException();
		}
	}

	protected int read(Buffer b,int size) throws SocketException
	{
		try {
			return peer_in.read(b.buffer,b.size,size);
		} catch (IOException io) {
			not_exit=false;
			throw new SocketException();
		}
	}

	protected void write(int r) throws SocketException
	{
		try {
			peer_out.write(buffer,0,r);
			peer_out.flush();
		} catch (IOException io) {
			not_exit=false;
			throw new SocketException();
		}
	}

	protected void write(Buffer b) throws SocketException
	{
		try {
			peer_out.write(b.buffer,0,b.size);
			peer_out.flush();
		} catch (IOException io) {
			not_exit=false;
			throw new SocketException();
		}
	}

	private int internalReturn(IOThread t,int rval,byte status)
	{
		do {
			try {
				if (status != -1) {
					t.setStatus(status);
				}
				t.join();
				return rval;
			} catch (InterruptedException ie) { }
		} while (true);
	}

	private void printPercentage(double total,long file_size)
	{
		try {
			int index;
			String perc=String.valueOf(total*100/file_size);

			index=perc.indexOf('.');
			if (index < 0 || index >= 3) {
				index=perc.length();
			} else {
				index+=3;
				if (index > perc.length()) {
					index=perc.length();
				}
			}
			screenMessage('\r');
			screenMessage("%"+perc.substring(0,index));
		} catch (Exception e) { }
	}

	protected String [] splitCommand(String cmdStr)
	{
		cmdStr=cmdStr.trim().replaceAll("\\\\[ ]","\0").replaceAll("\\\\[\t]","\1").replaceAll("[ \\t]{2,}"," ");
		String [] cmd=cmdStr.split(" ",-2);
		for (int i=0; i < cmd.length; i++) {
			cmd[i]=cmd[i].replaceAll("\0"," ").replaceAll("\1","\t");
		}
		return cmd;
	}

	protected void mkdir()
	{
		byte response=(byte)0;
		try {
			File f=getAbsoluteFile(reply);
			if (f.exists()) {
				if (f.isDirectory()) {
					response=(byte)1;
					local_pwd=f;
				}
			} else if (f.mkdirs()) {
				response=(byte)1;
				local_pwd=f;
			}
		} catch (Exception e) {
			screenMessage("Beklenmedik Durum: "+e.getMessage());
		}
		try {
			writeByte(response);
		} catch (SocketException se) { }
	}

	protected int downloadFile(File file) throws FileNotFoundException, SecurityException, NullPointerException, SocketException
	{
		int r_bytes,rb,chunk;
		long file_size;
		double total;

		if (file.exists() && (!file.isFile() || not_yes_all && (no_to_all || getUserChoice(file.getAbsolutePath()) <= (byte)1))) {
			writeByte(0);
			return FM_DOWNLOAD_CANCEL;
		} else {
			writeByte(1);
		}
		r_bytes=0;
		total=.0;
		screenMessage("Indirilen dosya: "+file.getAbsolutePath());
		screenMessage('\n');
		file_size=readLong();
		if (file_size >= 0) {
			BufferedFileWriter writer=new BufferedFileWriter(file,5,socket.getSendBufferSize());
			writer.start();
			Buffer b=writer.getBuffer();
			while (b != null && (r_bytes=readInt()) > 0) {
				do {
					chunk= r_bytes > (b.buffer.length - b.size) ? b.buffer.length - b.size : r_bytes;
					r_bytes-=chunk;
					while ((rb=read(b,chunk)) >= 0 && (chunk-=rb) > 0) {
						total+=rb;
						b.size+=rb;
						printPercentage(total,file_size);
					}
					if (rb >= 0) {
						total+=rb;
						printPercentage(total,file_size);
						if ((b.size+=rb) == b.buffer.length) {
							if (!writer.putBuffer(b)) {
								writeByte(0);
								return internalReturn(writer,FM_DOWNLOAD_CANCEL,BufferedFileWriter.BFW_STOPPED);
							}
							if ((b=writer.getBuffer()) == null) {
								break;
							}
						}
					} else {
						break;
					}
				} while (r_bytes > 0);
				if (b != null) {
					if (rb >= 0) {
						writeByte(1);
					} else {
						throw new SocketException();
					}
				}
			}
			if (b == null) {
				writeByte(0);
				return internalReturn(writer,FM_DOWNLOAD_CANCEL,BufferedFileWriter.BFW_STOPPED);
			}
			if (b.size > 0 && !writer.putBuffer(b)) {
				writeByte(0);
				return internalReturn(writer,FM_DOWNLOAD_CANCEL,BufferedFileWriter.BFW_STOPPED);
			}
			if (file_size == 0) {
				screenMessage("%100");
			}
			screenMessage('\n');
			if (!((b=writer.getBuffer()) != null && writer.putBuffer(b))) {
				writeByte(0);
				return internalReturn(writer,FM_DOWNLOAD_CANCEL,BufferedFileWriter.BFW_STOPPED);
			}
			if (r_bytes == 0) {
				writeByte(2);
				return internalReturn(writer,FM_DOWNLOAD_OK,(byte)-1);
			}
		}
		return FM_DOWNLOAD_CANCEL;
	}

	protected int uploadFile(File file) throws FileNotFoundException, NullPointerException,SecurityException, SocketException
	{
		int response;
		long file_size;
		double total;

		switch(readByte()) {
			case 0:
				return FM_UPLOAD_CANCEL;
			case 1:
				break;
			case 4:
				writeByte(getUserChoice(file.getAbsolutePath()));
				if (readByte() != (byte)1) {
					return FM_UPLOAD_CANCEL;
				}
				break;
			default:
				return FM_UPLOAD_INVALID;
		}
		screenMessage("Yuklenen dosya: "+file.getAbsolutePath());
		screenMessage('\n');
		BufferedFileReader reader=new BufferedFileReader(file,5,socket.getSendBufferSize());
		reader.start();
		response=1;
		file_size=file.length();
		total=.0;
		writeLong(file_size);
		try {
			Buffer b;
			while ((b=reader.getBuffer()) != null && b.size > 0) {
			 	writeInt(b.size);
				write(b);
				if ((response=readByte()) != 1) {
					break;
				}
				if (!reader.putBuffer(b)) {
					break;
				}
				total+=b.size;
				printPercentage(total,file_size);
			}
			if (file_size == 0) {
				screenMessage("%100");
			}
			screenMessage('\n');
			if (response == 1) {
				if (reader.getStatus() == BufferedFileReader.BFR_EOF) {
					writeInt(0);
					if (readByte() == 2) {
						return internalReturn(reader,FM_UPLOAD_OK,(byte)-1);
					}
				} else {
					writeInt(-1);
				}
			}
		} catch (IOException io) {
			writeInt(-1);
		}
		return internalReturn(reader,FM_UPLOAD_CANCEL,BufferedFileReader.BFR_STOPPED);
	}
}
